// client.c - Socket server command line client
//
// Makes a single connection to the server, sends the command, and prints the
// result.
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA

#include "common.h"
#include "util.h"
#include "server.h"
#include "conf.h"
#include "client.h"
#include "password.h"

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

int clientsocket;

// Avoid those annoying TCP_WAIT states
void cli_cleanup(void)
{
	if (clientsocket) close(clientsocket);
}

// Open a TCP connection to the server
// Returns 0 on all-OK
int cli_serverconnect(char *servername, int port)
{
	struct protoent *pe = NULL; // initialise to avoid warning
	struct sockaddr_in addr_in;
	struct sockaddr_un addr_un;
	struct hostent *host;
	int rc;
	int bytes;
	char *error;
	static int init;

	if (!init) { atexit(cli_cleanup); init=1; }
	error = NULL;
	if (port) {
		// TCP connect
		memset(&addr_in,0,sizeof(addr_in));
		pe = getprotobyname("tcp");
		if (!pe) { error="getprotobyname"; goto oops; }
		clientsocket = socket(AF_INET,SOCK_STREAM,pe->p_proto);
		if (clientsocket==-1) { error="socket"; goto oops; }
		host = gethostbyname(servername);
		if (!host) { error="gethostbyname"; goto oops; }
		addr_in.sin_port = htons(port);
		addr_in.sin_family = AF_INET;
		addr_in.sin_addr = *( (struct in_addr *)(host->h_addr_list[0]));
		rc = connect(clientsocket, (struct sockaddr *)&addr_in, sizeof(addr_in));
	}
	else {
		// Unix socket connect
		memset(&addr_un,0,sizeof(addr_un));
		clientsocket = socket(PF_UNIX,SOCK_STREAM,0);
		if (clientsocket==-1) { error="socket"; goto oops; }
		addr_un.sun_family = AF_UNIX;
		strncpy( addr_un.sun_path, servername, sizeof(addr_un.sun_path) );
		rc = connect(clientsocket, (struct sockaddr *)&addr_un, sizeof(addr_un));
	}

	// We should really try from **h_addr_list
	if (rc) { error="connect"; goto oops; }
	bytes = 1;
	if (port)
		setsockopt(clientsocket,pe->p_proto, TCP_NODELAY, &bytes, sizeof(bytes));
	return 0;

oops:
	perror(error);
	if (clientsocket) close(clientsocket);
	clientsocket = 0;
	return 1;
}


// A single-shot query to the server.  Not reentrant.
// Returns 0 on all-OK
int cli_serverquery(char *request, char *response, int responselen)
{
	char *error;
	char *estring, *p, *eol;
	int received;
	int requestlen;
	int bytes;
	int requests;

	// Write request
	requestlen=strlen(request);
	for (p=request, requests=1; p<request+requestlen; p++) {
		if (*p == '\n') requests++;
	}
	p = request;
	// With a trailing  \n
	estring = p+requestlen;
	*estring='\n';
	requestlen++;
	
	while (requestlen) {
		bytes = send(clientsocket, p, requestlen, MSG_NOSIGNAL);
		if (bytes<0 && errno==EINTR) bytes=0;
		else if (bytes<0) { *estring = 0; error="send"; goto oops; }
		requestlen -= bytes;
	}
	*estring = 0;

	// Read response
	received = 0;
	memset(response,0,responselen);
	responselen--;  // Make sure we keep the trailing null
	p = response;
	while (received < responselen) {
		bytes = recv(clientsocket, response+received, responselen-received, MSG_NOSIGNAL);
		if (bytes==0) break;
		if (bytes<0 && errno==EINTR) bytes=0;
		else if (bytes<0) { perror("recv"); return 1; }
		received += bytes;
		// Search response buffer for a blank line
		p=response;
		while (*p) {
			// At the end of a line?
			if (*p=='\r' || *p=='\n') {
				// FIXME: We should not spew the crlf at the end of the line
				if (--requests == 0) {
					*p=0;
					goto we_have_a_winner;
				}
			}
			eol = strchr(p,'\n');
			if ( !eol ) break;  // The line MUST be terminated
			p = eol + 1;
		}
	}

we_have_a_winner:
	return 0;

oops:
	if (error) perror(error);
	return 1;
}


// Close the connection to the server
// Returns 0 on all-OK
int cli_serverclose(void)
{
	if ( clientsocket && close(clientsocket) ) {
		perror("close");
		clientsocket=0;
		return 1;
	}
	clientsocket=0;
	return 0;
}


// Single-shot question and answer - -returns 0 on ok
int cli_askserver(char *servername, int port, char *request, char *response, int responselen)
{
	if (cli_serverconnect(servername,port)) goto oops;
	if (cli_serverquery(request,response,responselen)) goto oops;
	if (cli_serverclose()) goto oops;
	return 0;

oops:
	cli_serverclose();
	return 1;
}


// Be a client -- command line version
int cli_docommandline(int argc, char *argv[])
{
	char request[SOCK_INBUFFER];
	char response[SOCK_OUTBUFFER];
	int a;
	int rc;

	*request=0;
	*response=0;
	if (conf.autoauth) {
		rc=pass_getauthinfo(getuid(), request, sizeof(request),
			"UID ","\n");
	}
	for (a=1; a<argc; a++) {
		if (a!=1) strncat(request," ",sizeof(request)-strlen(request)-1);
		strncat(request,argv[a],sizeof(request)-strlen(request)-1);
	}
	// rc = cli_askserver("localhost",2200,request,response,sizeof(response));
	rc = cli_askserver(conf.unixsocket,0,request,response,sizeof(response));
	if (!rc) {
		printf("%s\n",response);
	}
	return rc;
}

// Send the command and the current user name to the server.  Print the answer.
int cli_usercommand(char *command)
{
	char request[SOCK_INBUFFER];
	char response[SOCK_OUTBUFFER];
	int rc;
	char authinfo[255];

	*response=0;
	authinfo[0] = 0;
	if (conf.autoauth) {
		rc=pass_getauthinfo(getuid(), authinfo, sizeof(authinfo),
			"UID ","\n");
	}
	snprintf(request,sizeof(request)-1,"%s%s",
		authinfo,
		command);
	rc = cli_askserver(conf.unixsocket,0,request,response,sizeof(response));
	if (!rc) {
		printf("%s\n",response);
	}
	return rc;
}

