// dialer.c
//
// Code to handle the execution of commands that do dialling and link
// monitoring.  Okay, so it's not really a dialer anymore, but hey.
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA

#include "common.h"

#include "shell.h"
#include "dialer.h"
#include "conf.h"

#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif


// Compile the regular expression, or panic
void regcomp_or_bust(regex_t *preg, const char *regex)
{
	int errcode;
	char errbuf[80];

	errcode = regcomp(preg, regex, REG_ICASE | REG_EXTENDED);
	if (errcode) {
		regerror(errcode, preg, errbuf, sizeof(errbuf));
		fprintf(stderr,"regex: %s\n",errbuf);
		exit(3);
	}
}

#if 0
// Allocate a new regular expression for the given regular expression
// values \1 \2 \3 \4 .. \9 are substituted with the parameter values before
// the regular expression is compiled.  If you include regular expression
// chars (e.g. brackets) in the parameter for now, you are in trouble.
regex_t *dlr_buildlinkregex(const char *original, char *param[])
{
	char regex[256];
	const char *p;
	char *q, *source;
	int n, paramcount ;

	int errcode;
	regex_t *preg;

	// Do string substitution.  We sure hope param is NULL terminated
	q = regex;
	paramcount=0;
	while (param[paramcount]) { paramcount++; }
	for (p=original; *p; p++) {
		if ( *p=='\\' && isdigit(*(p+1)) ) {
			p++;
			n = (*p - '1');   // \1 .. \9 => 0..8
			if ( n >= paramcount ) continue;
			source = param[n];
			while (*source) {
				*q++ = *source++;
				if (q==regex+sizeof(regex))
					return NULL;  // It's too long!
			}
		}
		else if ( *p=='\\' && *(p+1)=='\\' ) {
			*q++=*p++;
		}
		else *q++ = *p;
		if (q==regex+sizeof(regex))
			return NULL;  // It's too long!
	}
	*q=0;  // Null terminated string

	// Now compile it into a regular expression
	errcode = regcomp(preg, regex, REG_ICASE | REG_EXTENDED);
	if (errcode) {
		// Don't panic
		return NULL;  // Bad regular expression
	}
	return preg;

}
#endif

// Compiled regular expressions
regex_t reg_up,
	reg_notup,
	reg_rx,
	reg_tx,
	reg_islocal,
	reg_linkisvalid,
	reg_simplifyurl;

void dlr_cleanup(void)
{
	regfree(&reg_up);
	regfree(&reg_notup);
	regfree(&reg_rx);
	regfree(&reg_tx);
	regfree(&reg_islocal);
}

// Initialise regular expressions
void dlr_init(void)
{
	static int init;

	if (init) return;
	init = 1;
	regcomp_or_bust(&reg_rx, conf.reg_rx);
	regcomp_or_bust(&reg_tx, conf.reg_tx);
	regcomp_or_bust(&reg_up, conf.reg_up);
	regcomp_or_bust(&reg_notup, conf.reg_notup);
	regcomp_or_bust(&reg_islocal, conf.reg_islocal);
	regcomp_or_bust(&reg_simplifyurl, conf.reg_simplifyurl);
	regcomp_or_bust(&reg_linkisvalid, conf.reg_linkisvalid);
	atexit(dlr_cleanup);  // I wonder what happens if you free uninitalised regexes?
}


// Check whether a url is local or remote
int dlr_urlisremote(char *url)
{
	regmatch_t pmatch[1];

	dlr_init();
	return ( regexec(&reg_islocal, url, 1, pmatch, 0) != 0 ); // True == does not match locally 
}

// Check whether a named link is valid
int dlr_linkisvalid(char *linkname)
{
	regmatch_t pmatch[1];

	dlr_init();
	return ( regexec(&reg_linkisvalid, linkname, 1, pmatch, 0) == 0 );
	// Return valid of true == link is valid
}

// Count the packets
// Note to self about regex matches:
// The 0'th match is the entire string, the 1'th match is the optional part, and
// the 2nd part is the site...
char *dlr_simplifyurl(char *url)
{
	int offset,len;
	regmatch_t pmatch[3];

	dlr_init();  // initialise regular expressions

	// Check for up link:
	if (regexec(&reg_simplifyurl, url, 3, pmatch, 0)) 
		return url; // oops!  No match!
	offset = pmatch[2].rm_so;
	len = pmatch[2].rm_eo - offset;
	memmove(url,url+offset,len);
	url[len]=0;
	return url;
}

// Count the packets
// Note to self about regex matches:
// The 0'th match is the entire string, the 1'th match is the
// first substring
int dlr_packetstats(int *ifup, int *rxpackets, int *txpackets) 
{
	int len;
	char buffer[500];
	regmatch_t pmatch[3];

	dlr_init();  // initialise regular expressions

	len = sizeof(buffer);
	buffer[0]=0;
	shell_inline(conf.cmdcheck,buffer,&len);
	*ifup = 1;
	// Check for up link:
	if (regexec(&reg_up, buffer, 3, pmatch, 0)) 
		*ifup = 0; // No match?  Link is down!
	if (regexec(&reg_notup, buffer, 3, pmatch, 0)==0) 
		*ifup = 0; // No, we shouldn't have matched this
	if (!ifup)
		return 0;

	// Run the command to count packets if it is different
	if (strcmp(conf.cmdpackets,conf.cmdcheck)!=0) {
		len = sizeof(buffer);
		buffer[0]=0;
		shell_inline(conf.cmdpackets,buffer,&len);
	}
	// Count TX packets
	if (regexec(&reg_rx, buffer, 3, pmatch, 0)) 
		return -1; // oops!  No match!
	*rxpackets = atoi(&buffer[pmatch[1].rm_so]);
	// Count RX packets
	if (regexec(&reg_tx, buffer, 3, pmatch, 0)) 
		return -1; // oops!  No match!
	*txpackets = atoi(&buffer[pmatch[1].rm_so]);
	return 0;
}
// Find (all) the matching strings for a given regular expression
// Return an array of (1) pointers to those strings, and
//	(2) lengths of those strings.
// Note again to self:
// The 0'th match is the entire string, the 1'th match is the optional part, and

char *dlr_getmatchstring(
	regex_t *ureg,	// somewhere in the past we compiled the regex
	char *data,	// Here's what we are searching through (null terminated)
	int *matchlen,	// Return: length of matching string
	int matchindex)	// Howmanyeth string to match (0 is the whole bang shoot, but you can't have it)
{
	int offset;
#define MAX_PMATCH_TERMS 10
	regmatch_t pmatch[MAX_PMATCH_TERMS];

	if (!ureg || !data || !matchlen ||
		matchindex<0 || matchindex >= MAX_PMATCH_TERMS ) {
		return NULL;  // be nice or we won't talk to you
	}

	// Find the matching part(s)
	if (regexec(ureg, data, MAX_PMATCH_TERMS, pmatch, 0)==0) {
		return NULL; // no match
	}
	offset = pmatch[matchindex].rm_so;
	*matchlen = pmatch[matchindex].rm_eo - offset;
	return &data[offset];
}

