// logic.c
//
// Handle socket events (this is the main control module for dialserver)
// If ever there was a misnamed piece of code, this could be it :)
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA

#include "common.h"

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_REGEX_H
#include <regex.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include "password.h"
#include "movingavg.h"
#include "log.h"
#include "shell.h"
#include "logic.h"
#include "dialer.h"
#include "conf.h"
#include "blah.h"

// --------------------------------------------------------------------
// --  Dialling server
// --------------------------------------------------------------------

#define MAX_LINKS 64

#define FLAG_AUTH 1

// What states exist
typedef enum _DIALSTATE { 
	OFFLINE, 
	DIALLING, 
	DIALLED, 
	HANGUP } DIALSTATE;

// State of the dial in link
typedef struct _DIALINFO {
	DIALSTATE state; 	// What state are we in now
	time_t started;		// When did we start the dial-up
	time_t active;		// When did we detect the line being up
	time_t hangup;		// When did we start hanging up
	time_t hungup;		// When did we finish hanging up
	int rxpackets;		// Last checked rx packets
	int txpackets;		// Last checked tx packets
	time_t refresh;		// When must we wake up to send and receive mail
	time_t wakeoff;		// When must we wake up to check for queued mail
	time_t idlecheck;	// Next scheduled cmdlistidle
	int failurecount;	// How many times have we failed to dial
	time_t lastcall;	// When did we last start a call
	int minnotified;	// Have we sent a "not hanging up" message?
	int alllogedout;	// Set when everyone has logged out
	char reason[64];	// How did we reach the state we are in?
	pid_t pidduring;	// PID of command run 'during' dialout
	pid_t pidhangup;	// PID of command for hanging up the dialout
	pid_t pidrefresh;	// PID of the refresh command (sendreceivemail)
	MOVINGAVERAGE txavg;	// We're watching what you're sending
	MOVINGAVERAGE rxavg;	// ... and receiving
} DIALINFO;

// State of a connected link
typedef struct _LINKDATA {
	char owner[32];		// The users name, if we have it -- otherwise IP address
	char ipaddress[32];	// IP address as a string (this can be 'mail' or something)
	time_t started;		// Time at which this guy joined
	time_t lastevent;	// Time at which this guy did something
	int idledetect;		// Did we detect idleness?
	regex_t reg_txlink;	// Get transmitted data
	regex_t reg_rxlink;	// Get received data
	MOVINGAVERAGE txdata;	// Transmitted data
	MOVINGAVERAGE rxdata;	// Received count
} LINKDATA;

time_t server_start_time;
DIALINFO dial;
LINKDATA linkdata[MAX_LINKS];
int dlr_sighup = 0;
int dlr_sigusr1 = 0;
int dlr_sigusr2 = 0;

// Alpha - make time go smoothly, even if the clock is jumpy
// FIXME! UNTESTED
time_t xtime(time_t *atime)
{
	static time_t lasttime;
	static int offset;
	time_t thistime;
	thistime=time(atime);
	if (thistime != -1) {
		if (lasttime>thistime) {
			// Time went backwards - make it move by 5 second!
			offset+=lasttime-thistime+5;
		}
		// 5 minute interval must become 5 minutes
		else if (lasttime+300<thistime) { 
			offset+=lasttime-thistime-300;
		}
		lasttime=thistime;
		if (atime != NULL) {
			*atime += offset;
		}
	}
	return thistime+offset;
}

// Child reaper
void srv_commanddone(pid_t pid)
{
	time_t thetime;

	if (pid==dial.pidduring) {
		dial.pidduring = 0;
	}
	else if (pid==dial.pidhangup) {
		dial.pidhangup = 0;
	}
	// Complete sendreceivemail script and schedule the next one ...
	else if (pid==dial.pidrefresh) {
		dial.pidrefresh = 0;
		dial.refresh = time(&thetime) + conf.timerefresh;
	}
}

// See who they said they were and decide whether that is okay
// Result: fill in userdata with supplied user name if okay
int dlr_authuser(char *param, char *userdata, int sizeofuserdata)
{
	char *username, *password;
	char *p;

	p = param;
	while (*p && isspace(*p)) { p++; }
	username=strtok(p," \t\r\n");
	password=strtok(NULL,"\n");
	if (!username || !password)
		return 0; // Sorry
	// FIXME : User x, pass x
	if (pass_checkauth(username,password)) {
		strncpy(userdata,username,sizeofuserdata-1);
		userdata[sizeofuserdata-1]=0;
		return 1;
	}
	// sorry
	return 0;
}

// See who they said they were and decide whether that is okay
// Result: fill in userdata with supplied user name if okay
int dlr_authuid(char *param, char *userdata, int sizeofuserdata)
{
	char realkey[100];
	char *p;
	uid_t uid;
	struct passwd *pwd;
	int rc;

	p = param;
	while (*p && isspace(*p)) { p++; }
	uid = atoi(p);
	*realkey=0;
	rc=pass_getauthinfo(
		uid,
		realkey,
		sizeof(realkey)-1,"","");
	//Authorisation debugging:
	//fprintf(stderr," Real: %s\nGiven: %s\n",
	//	realkey, param);
	if ((rc==0) && (strncmp(realkey,param,strlen(realkey))==0) ) {
		pwd = getpwuid( uid );
		strncpy(userdata,pwd->pw_name,sizeofuserdata-1);
		userdata[sizeofuserdata-1]=0;
		return 1;
	}
	// Notice how we don't print out an error message ...
	return 0; // Sorry
}

// Find an existing link record for an IP address
// Returns 1 on failure, 0 on success and sets the result
// ipaddress: the link we are searching for
// client: the connection we are doing this for
// createnew: create the link if it is not active
LINKDATA *dlr_findlink(char *ipaddress, CLIENTBUFFER *client, int createnew)
{
	int a;
	int avail;

	// Deal with link-name locks rather than IP-addresses
	if (!ipaddress || !ipaddress[0])
		return NULL; // ah! get lost
	
	// Find a link monitoring slot 
	avail=0;
	for (a=0; a<MAX_LINKS; a++) {
		if (strcmp(linkdata[a].ipaddress,ipaddress)==0 ) {
			return &linkdata[a];
		}
		if (!avail && !linkdata[a].started)
			avail=a;
	}
	// Create new link slot
	if (avail && createnew) {
		if (!dlr_linkisvalid(ipaddress))
			return NULL;
			/* Improperly formatted link destination */
		memset(&linkdata[avail], 0, sizeof(linkdata[avail]));

		// Now, find out who they are, and whether they are allowed in or not
		// If they are not allowed in, then the output must be nothing from
		// ip to name script
		// Is it a real IP address?
		strncpy(linkdata[avail].ipaddress,ipaddress,sizeof(linkdata[avail].ipaddress)-1);
		strncpy(linkdata[avail].owner,client->userdata, sizeof(linkdata[avail].owner)-1);
	}
	else {
		// FIXME -- Reached link quota -- oops!
		return NULL;
	}
	return &linkdata[avail];
}

// Returns 1 if all users registered have disconnected
int dlr_countlinks(void)
{
	int count = 0;
	int a;

	for (a=0; a<MAX_LINKS; a++) 
		if (linkdata[a].started) count++;
	return count;
}

void dlr_logoutlink(LINKDATA *linkdata, char *reason);

// Disconnect all clients
void dlr_disconnectalllinks(void)
{
	int a;
	for (a=0; a<MAX_LINKS; a++) {
		if (linkdata[a].started) {
			dlr_logoutlink(&linkdata[a],"dropped");
		}
	}
	memset(linkdata,0,sizeof(linkdata));
}

// Disconnect one client, if active
int dlr_disconnectlink(char *linkname, char *reason)
{
	int a,count;
	count=0;
	for (a=0; a<MAX_LINKS; a++) {
		if (strcmp(linkdata[a].ipaddress,linkname)==0) {
			count++;
			dlr_logoutlink(&linkdata[a],reason);
		}
	}
	return count;
}

// Run the idle check, and update the lastevent value for all non-idle links.
// Disconnect all links that have timed out
void dlr_cmdlistidle(time_t thetime)
{
	char *idlelist, *next, *ipaddress;
	int exitstatus;
	int a;

	// Run the command and mark idle links as idledetect==1;
	idlelist = shell_inline_alloc(conf.cmdlistidle, &exitstatus);
	if (idlelist) {
		// Before we start, nothing is idle
		for (a=0; a<MAX_LINKS; a++) { linkdata[a].idledetect = 0; }

		next=idlelist;
		while ( (ipaddress = strtok(next," \t\r\n")) ) {
			next=NULL;
			for (a=0; a<MAX_LINKS; a++) {
				if (strcmp(linkdata[a].ipaddress,ipaddress)==0 ) {
					linkdata[a].idledetect = 1;
				}
			}
		}
		free(idlelist);

		// Only run this if we managed to list idle links ...
		// If the link aint idle .. it's busy, so note the event
		// If it's idle, log it out.
		for (a=0; a<MAX_LINKS; a++) {
			if ( linkdata[a].idledetect==0 ) {
				linkdata[a].lastevent = thetime;
			}
			else if (thetime >= (linkdata[a].lastevent + conf.timelinkidle)) {
				dlr_logoutlink(&linkdata[a],"idle");
			}
		}
	}
}

// Script that gets run every so often when we're connected
void dlr_dorefresh(time_t thetime)
{
	// dlr_event("Refresh script\n");
	// Only run one-instance of the script:
	if (dial.pidrefresh == 0) { 
		dial.pidrefresh = shell_exec_bg("cmdrefresh",conf.cmdrefresh);
		dial.refresh = thetime + conf.timerefresh;
	}
}

// Do the appropriate client things after a switch to a new state has been
// detected or initated
void dlr_newstate(int state)
{
	int dtime;
	int lineup;
	int txpackets, rxpackets;

	if (dial.state==state) return;

	// Changing state?  First kill the 'during' script
	if (dial.pidduring) {
		// Send a hangup to the shell process, which will hopefully clean up
		kill(dial.pidduring,SIGHUP);
		dial.pidduring=0;
	}

	switch (state) {
	case HANGUP:
		// Disconnect everybody - the ship is leaving
		dlr_disconnectalllinks();
		time(&dial.hangup);
		dial.wakeoff = 0;
		// If we were busy running the refresh script, cancel it
		if (dial.pidrefresh) {
			kill(dial.pidrefresh,SIGHUP);
		}
		break;
	case DIALLING:
		time(&dial.started);	// a connection is now established
		dial.reason[0] = 0;	// No reason for hanging up - we're on line
		dial.state = DIALLING;
		dial.alllogedout = 0;
		dial.wakeoff = 0;
		break;
	case DIALLED:
		dial.failurecount = 0;  // Success did happen at some point
		time(&dial.active);
		dial.lastcall = dial.active;
		dial.minnotified = 0;
		shell_exec_bg("afterdial",conf.cmdafterdial);
		dlr_dorefresh(dial.active);
		// Find initial packet count -- for ISDN, the initial
		// packet count is the packet count of the interface,
		// not the link...
		dlr_packetstats(&lineup, &rxpackets, &txpackets);
		// Record the first packet count in the moving buffer thing
		mva_start(&dial.rxavg,rxpackets);
		mva_start(&dial.txavg,txpackets);
		child_notify = srv_commanddone; 
		dial.pidduring = shell_exec_bg("during",conf.cmdduring);
		dial.wakeoff = 0;
		break;
	case OFFLINE:
		// The connection, she is dead
		time(&dial.hungup);
		dtime = dial.hungup - dial.started;
		dlr_event("call-duration %d.%.2d (%s)\n",
			(int)(dtime/60), (int)(dtime%60),
			dial.reason);
		// If we were busy running the refresh script, cancel it
		if (dial.pidrefresh) {
			kill(dial.pidrefresh,SIGHUP);
		}
		// And schedule next check for on-line-ness
		dial.wakeoff = dial.hungup + conf.timewakeoff;
		break;
	}
	dial.state=state;
}

// Script that gets run every so often when we're not dialled in - just to see
// if we should be (for example, if there is unsent mail in the queue, it's
// nice to dial)
void dlr_dowakeoff(time_t thetime)
{
	char number[40];
	time_t t, dtime;

	dtime = time(&t) - dial.hungup;
	snprintf(number,sizeof(number),"%d", (int)dtime);
	setenv("dsSECONDS",number,1);
	snprintf(number,sizeof(number),"%d", (int)(dtime/60));
	setenv("dsMINUTES",number,1);
	// dlr_event("Off-line check script\n");
	shell_exec_bg("cmdwakeoff",conf.cmdwakeoff);
	dial.wakeoff = thetime + conf.timewakeoff;
}

// Check the dialling state - and make all sorts of interesting remarks about it
int dlr_checkstate(void)
{
	static time_t lastcheck;
	time_t thetime;
	int lineup;
	int pkcount;
	int txpackets, rxpackets;

	// Don't hammer that system command
	if (abs(time(&thetime)-lastcheck)<conf.timepoll) 
		return 1;
	// Round the time down -- the first beat will be shorter, but the
	// remaining beats will be regular.
	lastcheck=thetime - (thetime % conf.timepoll);

	dlr_packetstats(&lineup, &rxpackets, &txpackets);

	switch (dial.state) {
	case DIALLED:
		if (!lineup) {
			snprintf(dial.reason,sizeof(dial.reason)-1,
				"Connection lost unexpectedly");
			// dlr_event("%s\n",dial.reason);
			dlr_newstate(HANGUP);
			shell_exec_bg("afterhangup",conf.cmdafterhangup);
			dlr_newstate(OFFLINE);
		} else {
			// Idle link detection
			// THe link is idle if --
			//   There are not that many packets
			//   There are no active links, and the line has been up longer than 3 minutes
			// Do a packet count, and log it to the moving history
			mva_addvalue(&dial.rxavg,rxpackets);
			mva_addvalue(&dial.txavg,txpackets);
			// Check the packet count for the last while ...
			pkcount = mva_integral(&dial.txavg,conf.timepacketcheck/conf.timepoll,INTEGRAL_STRICT);

			if (thetime > dial.idlecheck) {
				// Check for idle links, and schedule next check
				dlr_cmdlistidle(thetime);
				dial.idlecheck+=conf.timeidlecheck;
				if (dial.idlecheck<=thetime) {
					dial.idlecheck=thetime+conf.timeidlecheck;
				}
			}
			// Link idle because all logged out a little while ago?
			if (dial.alllogedout && thetime - dial.started >= conf.timeminimumcall) {
				// dlr_event("%s\n",dial.reason);
				dlr_hangupnow(NULL,"No links active");
			}
			// Link idle because it's not that busy?
			else if (pkcount>=0 && pkcount<conf.minimumpackets) {
				if (thetime - dial.lastcall < conf.timeminimumcall) {
					// Don't hang up if we have just started a call, but
					// do explain.
					if (!dial.minnotified) 
						dlr_event("Not hanging up idle link yet, %d seconds remain\n",
							conf.timeminimumcall - (thetime - dial.lastcall) );	
					dial.minnotified = 1;
				}
				else {
					// Hang up idle link
					snprintf(dial.reason,sizeof(dial.reason)-1,
						"Only %d of %d packets in %d seconds",
						pkcount, conf.minimumpackets, 
						conf.timepacketcheck/conf.timepoll*conf.timepoll);
					dlr_hangupnow(NULL,dial.reason);
				}
			}
			else {
				// We didn't just hang up, so let's see if we
				// want to generate some traffic
				if (thetime >= dial.refresh) {
					dlr_dorefresh(thetime);
				}
			}
		}
		break;
	case DIALLING:
		if (lineup) {
			// dlr_event("Dialling successful\n"); // yawn
			dlr_newstate(DIALLED);
		}
		else if (thetime-dial.started > conf.timeoutdial) {
			char number[12];
			snprintf(dial.reason,sizeof(dial.reason)-1,
				"Dialling timed out after %d seconds",
				conf.timeoutdial);
			shell_exec_bg("hangup",conf.cmdhangup);
			snprintf(number,sizeof(number),"%d", ++dial.failurecount);
			setenv("dsFAILURECOUNT",number,1);
			child_notify = srv_commanddone; 
			dial.pidhangup = shell_exec_bg("dialtimeout", conf.cmddialtimeout);
			dlr_newstate(HANGUP);
			if (dial.pidhangup==0) {
				dlr_newstate(OFFLINE);
			}
		}
		break;
	case HANGUP:
		// We insist that the hangup be complete before we pronounce the hangup complete
		if (dial.pidhangup==0 && lineup==0) {
			shell_exec_bg("afterhangup",conf.cmdafterhangup);
			dlr_newstate(OFFLINE);
		}
		else if (thetime-dial.hangup > conf.timeouthangup) {
			dlr_event("Hangup timed out after %s seconds\n",conf.timeouthangup);
			shell_exec_bg("hanguptimeout",conf.cmdhanguptimeout);
			shell_exec_bg("hangup",conf.cmdhangup);
			dial.pidhangup = 0;
			dlr_newstate(OFFLINE);
		}
		break;
	case OFFLINE:
		// We should really be more cool about this kind of thing
		if (lineup) {
			dlr_event("Connection established unexpectedly\n");
			dlr_newstate(DIALLING);
			dlr_newstate(DIALLED);
		}
		else {
			if (dial.wakeoff && thetime >= dial.wakeoff) {
				dlr_dowakeoff(thetime);
			}
		}
		break;
	}
	return 0;
}


// Adds a link to the logged in list
void dlr_loginuser(LINKDATA *linkdata)
{
	time_t now;
#if 0
	char *parameters[5];
#endif

	time(&now);
	dial.lastcall = now;
	dial.minnotified = 0;
	linkdata->started = now;
	linkdata->lastevent = now;
	dlr_event("start %s (%s)\n",linkdata->ipaddress,linkdata->owner);
	setenv("dsUSERIDENT",linkdata->owner,1);
	setenv("dsUSERIP",linkdata->ipaddress,1);
	shell_exec_bg("login",conf.cmdlogin);

#if 0
	// Set up regular expressions for measuring link activity
	parameters[0] = linkdata->ipaddress;
	parameters[1] = linkdata->owner;
	parameters[2] = NULL;
	linkdata->reg_txlink = dlr_buildlinkregex(conf.reg_linktx, &parameters);
	linkdata->reg_rxlink = dlr_buildlinkregex(conf.reg_linkrx, &parameters);
#endif
	// We don't start with a measurement, since that introduces an
	// unhealthy race condition between bringing the link up and watching
	// it's traffic.  Instead, we just pretend that this race condition
	// does not exist :)
	mva_start(&linkdata->txdata,0);
	mva_start(&linkdata->rxdata,0);
}


// Removes a link from the logged-in list
// Also makes a neat log entry about what they were doing on line
void dlr_logoutlink(LINKDATA *linkdata, char *reason)
{
	time_t now;
	time_t dtime,t;
	struct tm *local;
	char number[40];

	time(&now);
	dtime = time(&t) - linkdata->started;
	local = localtime(&linkdata->started);
	//dlr_event("start %s (%s)\n",linkdata->ipaddress,linkdata->owner);
	dlr_event("stop %s (%s) %d.%.2d %s\n",
		linkdata->ipaddress,
		linkdata->owner,
		(int)(dtime/60), (int)(dtime%60),
		reason);

	setenv("dsUSERIDENT",linkdata->owner,1);
	setenv("dsUSERIP",linkdata->ipaddress,1);

	snprintf(number,sizeof(number),"%d", (int)dtime);
	setenv("dsSECONDS",number,1);

	snprintf(number,sizeof(number),"%d.%.2d", (int)(dtime/60), (int)(dtime%60));
	setenv("dsTIME",number,1);

	shell_exec_bg("logout",conf.cmdlogout);
	// Bye bye
	memset(linkdata,0,sizeof(*linkdata));

	dial.alllogedout = (dlr_countlinks()==0);
}


	// Get the line status, and how long it has been like that...
char *dlr_getlinestatus(int *secs)
{
	time_t thetime;
	int a;

	if (!secs) secs = &a;  // Allow NULL
	time(&thetime);

	*secs = 0;
	switch (dial.state) {
	case DIALLING:
		*secs = thetime - dial.started;
		return "Dialling";
		break;
	case HANGUP:
		*secs = thetime - dial.hangup;
		return "Hanging up";
		break;
	case OFFLINE:
		*secs = ( thetime - dial.hungup) % (24*60*60);
		return "Off line";
		break;
	case DIALLED:
		*secs = ( thetime - dial.active) % (24*60*60);
		return "Dialled in";
		break;
	}
	return "";
}


// We want to dial because someone asked about it
// Returns 1 if we dialled in (and the user should be logged in)
int dlr_dialnow( CLIENTBUFFER *client, char *reason )
{
	int secs; 
	int result;
	char *linestatus;

	switch (dial.state) {
	case DIALLED:
		result = 1;
		break;
	case DIALLING:
		result = 1;
		break;
	case HANGUP:
		// We wait until the hangup is complete before we respond to the dial message
		result = 0;
		break;
	case OFFLINE:
	default:
		dlr_event("Dialling (%s)\n",reason);
		shell_exec_bg("dial",conf.cmddial);
		dlr_newstate(DIALLING);
		result = 1;
		break;
	}

	linestatus = dlr_getlinestatus(&secs);
	srv_printf(client,"%s: %d.%.2d\r\n", linestatus, secs/60, secs%60);
	return result;
}


// Hang up the connection, because someone (everyone?) asked nicely
void dlr_hangupnow( CLIENTBUFFER *client, char *reason )
{
	int secs; 
	char *linestatus;

	if (dial.reason!=reason) {
		snprintf(dial.reason,sizeof(dial.reason)-1, "%s", reason);
	}
	switch (dial.state) {
	case DIALLING:
		// FIXME -- the assumption is that we can abort a pending
		// dial by running hangup - this will happen if someone hits
		// dial, quickly followed by hangup.
		dlr_event("Interrupted dialling (%s)\n",reason);
	case DIALLED:
		dlr_newstate(HANGUP);
		dlr_event("Hanging up (%s)\n",reason);
		shell_exec_bg("hangup",conf.cmdhangup);
		break;
	case HANGUP:
		break;
	case OFFLINE:
	default:
		break;
	}

	linestatus = dlr_getlinestatus(&secs);
	srv_printf(client,"%s: %d.%.2d\r\n", linestatus, secs/60, secs%60);
}

// Show an activity graph
void dlr_showhistory( CLIENTBUFFER *client, int count )
{
	int max;
	int n;
	int a;
	MOVINGAVERAGE *avg;

	if (count==0) count=12;
	for (n=0; n<2; n++) {
		if (n==0) {
			srv_printf(client,"RX");
			avg = &dial.rxavg;
		}
		else {
			srv_printf(client,"TX");
			avg = &dial.txavg;
		}
		max = mva_count(avg)-1;
		if (count>max) count = max;
		for ( a=1 ; a<count; a++ ) {
			srv_printf(client," %d", mva_getdiff(avg,a,a+1));
		}
		srv_printf(client,"\r\n");
	}
	srv_printf(client,"%d seconds per interval\r\n", 
		(int)conf.timepoll);
}

// Run the diagnosis command and show the output
void dlr_showdiagnosis( CLIENTBUFFER *client )
{
	char buffer[4000]; // This size is random
	int bufsize = sizeof(buffer);
	char *p,lastchar=0;

	shell_inline(conf.cmddiagnose,buffer,&bufsize);
	if (bufsize<0) bufsize=0;
	buffer[bufsize]=0;
	while (buffer && buffer[bufsize-1]=='\n') {
		buffer[--bufsize]=0;
	}
	for (p=buffer; *p; p++) {
		if (lastchar=='\n' && *p=='\n') srv_putc(client, ' ');
		lastchar = *p;
		srv_putc(client, lastchar);
	}
	srv_printf(client,"\r\n");
}

// Show who is logged on, and what they have been doing lately
//	. client 127.0.0.1 00:21:00 1:10 http://www.squid.org:80/
//	. client 10.0.0.22 00:10:10 0:15 http://www.exotic.com/fish
//	. client 10.0.0.23 00:20:10 2:11 http://www.relax.com/comics
void dlr_showwho( CLIENTBUFFER *client )
{
	int a;
	int any;
	time_t thetime;
	int secs, secs2;

	time(&thetime);
	any = 0;
	for (a=0; a<MAX_LINKS; a++) if (linkdata[a].started) {
		secs = thetime - linkdata[a].started;
		secs2 = thetime - linkdata[a].lastevent;
		srv_printf(client,"%s: authorised %s for %d.%.2d",
			linkdata[a].owner, 
			linkdata[a].ipaddress,
			secs / 60, secs % 60,
			secs2 / 60, secs2 % 60);
		if (secs2 > conf.timeidlecheck) {
			srv_printf(client," (idle %d.%.2d)", 
				secs2 / 60, secs2 % 60);
		}
		srv_printf(client,"\r\n");
		any = 1;
	}
	if (!any) {
		srv_printf(client,"nobody: 0.00 (0.00) nobody logged on\r\n");
	}
}


// Show the line status
//	. users 3, call-duration 00:21:00, idle 0:15 (if users!=0)
//	. transmitted 200 packets, 60kB, 23 bps	<just an idea>
//	. received 200 packets, 60kB, 23 bps	<just idea>
//	+ Dialled in
//	+ Dialling ... 2
//	+ Dialling ... 4
//	+ Link up 0:21
//	+ No answer
void dlr_showstatus( CLIENTBUFFER *client, LINKDATA *alink)
{
	int secs;
	time_t thetime;
	int a;
	int linkcount;
	int rxcount, txcount;
	char *linestatus;

	time(&thetime);

	if (alink) {
		if (alink->started) {
			secs = thetime - alink->started;
			srv_printf(client,"Link status: Active %d.%.2d\r\n", 
				secs/60, secs%60);
			srv_printf(client,"Link owner: %s\r\n", 
				alink->owner);
		} 
		else {
			srv_printf(client,"Link status: Not active\r\n");
		} 
	}

	linestatus = dlr_getlinestatus(&secs);
	srv_printf(client,"Line: %s %d.%.2d\r\n", linestatus, secs/60, secs%60);
	if (dial.reason[0])
		srv_printf(client,"Reason: %s\r\n", dial.reason);

	linkcount = dlr_countlinks();
	srv_printf(client,"Links: %d %s connected\r\n",
		linkcount, linkcount==1 ? "link" : "links");
	switch (dial.state) {
	case DIALLING:
		break;
	case HANGUP:
		break;
	case OFFLINE:
		break;
	case DIALLED:
		srv_printf(client,"Total traffic: RX %d TX %d\r\n", 
			mva_integral(&dial.rxavg,0,INTEGRAL_ALL),
			mva_integral(&dial.txavg,0,INTEGRAL_ALL));

		// Print the traffic over the last while
		for (a=1; a<=2; a++) {
			int ticks;
			ticks = conf.timepacketcheck/a/conf.timepoll;
			rxcount = mva_integral(&dial.rxavg,ticks,INTEGRAL_SLOPPY);
			txcount = mva_integral(&dial.txavg,ticks,INTEGRAL_SLOPPY);
			srv_printf(client,"Last %d seconds: RX %d TX %d",
				ticks*conf.timepoll,
				rxcount,
				txcount
				);
			if (a==1) {
				srv_printf(client," (minimum %d)\r\n",conf.minimumpackets);
			}
			else {
				srv_printf(client,"\r\n");
			}
		}
		break;
	}
}

void dlr_sigterm_handler(int signum)
{
	dlr_event("Terminating on SIGTERM\n");
	exit(-SIGTERM);
}

void dlr_sigint_handler(int signum)
{
	dlr_event("Terminating on SIGINT\n");
	exit(-SIGINT);
}

// killall -USR1 dialserver -- dial for mail check
void dlr_sigusr1_handler(int signum)
{
	dlr_sigusr1 = 1;
}

// killall -USR2 dialserver -- check for mail now
// This is not used by anything yet ... - if you use it, I won't take it away
// later ... tell me
void dlr_sigusr2_handler(int signum)
{
	dlr_sigusr2 = 1;
}

void dlr_sighup_handler(int signum)
{
	dlr_sighup = 1;
}

void dlr_serverexit(void)
{
	time_t dtime;
	time(&dtime);
	dtime -= server_start_time;
	dlr_disconnectalllinks();
	dlr_event("Server exiting after %d.%.2d\n",dtime/60,dtime%60);
}

void dlr_serverstart(void)
{
	dlr_event("Dialserver " VERSION " started\n");
	time(&server_start_time);
	dial.started  = server_start_time;
	dial.active   = server_start_time;
	dial.hangup   = server_start_time;
	dial.hungup   = server_start_time;
	dial.lastcall = server_start_time;
	dial.wakeoff  = server_start_time+conf.timewakeoff;
	signal(SIGTERM,dlr_sigterm_handler);
	signal(SIGINT,dlr_sigint_handler);
	signal(SIGHUP,dlr_sighup_handler);
	signal(SIGUSR1,dlr_sigusr1_handler);
	signal(SIGUSR2,dlr_sigusr2_handler);
}

void dlr_stoplink(time_t thetime,char *linkname, CLIENTBUFFER *client)
{
	LINKDATA *linkdata;

	linkdata = dlr_findlink(linkname, client, 0);
	if (!linkdata) {
		srv_printf(client,"Link %s not active\r\n",linkname);
	}
	else {
		// FIXME: We could check whether the person is
		// ALLOWED to close the link
		dlr_logoutlink(linkdata,"requested");
		if (dial.alllogedout && thetime-dial.started > conf.timeminimumcall) {
			dlr_hangupnow(client,"All links closed");
		}
		srv_printf(client,
			"Link %s closed for %s\r\n",
			linkname, client->userdata);
	}
}

void dlr_startlink(time_t thetime, char *linkname, CLIENTBUFFER *client)
{
	LINKDATA *linkdata;
	int secs;

	linkdata = dlr_findlink(linkname, client, 1);
	if (!linkdata) {
		srv_printf(client,"Cannot create link %s for %s. Please try again.\r\n",
		linkname,client->userdata);
	}
	else if (linkdata->started) {
		// Retarting existing link
		dial.alllogedout = 0; // oops, was missing
		secs = thetime - linkdata->started;
		srv_printf(client,"Link is active %d.%.2d (%s)\r\n",
			secs/60, secs%60, dlr_getlinestatus(NULL));
	}
	else {
		dial.alllogedout = 0; // oops, was missing
		if (dlr_dialnow(client,/*request by*/client->userdata)) { // dial and say something
			// And log the link in
			dlr_loginuser(linkdata);
		}
	}
}

// A server: 
// 	Accepts one line commands, responds with multi-line return.  Last line starts with a
//	+ (because it seemed like a nice idea at the time)
// Accepts the following commands:
//
// From WEB front-end:
//	HELP 		Returns the number of bytes in the input buffer
//	LOGIN <user> <pass> All we do from now we attest to with this user and password
//	START <ip>	Dial in, and mark the specified IP address as an active link
//			+ Dialling
//	LINK 		Idly check whether the link is up
//			+ Link up 0:02:02, 2 users (how long the link has been up)
//			- Link down 2:21:21  (how long the link has been down)
//	STATUS		Returns the list of active IP's, total time, idle time and 
//			most recent URL.
//			. users 3, call-duration 00:21:00, idle 0:15 (if users!=0)
//			. client 127.0.0.1 00:21:00 1:10 http://www.squid.org:80/
//			. client 10.0.0.22 00:10:10 0:15 http://www.exotic.com/fish
//			. client 10.0.0.23 00:20:10 2:11 http://www.relax.com/comics
//			. transmitted 200 packets, 60kB, 23 bps
//			. received 200 packets, 60kB, 23 bps
//			+ Dialled in
//			+ Dialling ... 2
//			+ Dialling ... 4
//			+ Link up 0:21
//			+ No answer
//	STATUS <ip>	Returns whether the specified IP address is active, and the state
//			of a connection which is being established, if any.
//	LOGOUT <ip>	Remove the specified IP address from the active link list
//			+ Disconnected, 2 users still on line 
//			+ Disconnected, hanging up
//	UID <uid> <key> Set the user name to the unix user name of the uid
//			given, provided the key is stored in their key file
// From SQUID:
//	AUTH url ip-address/fqdn ident
//			Checks whether the IP address is in the active list.  If the IP
//			address is active, the usage information is updated.
//			+ IP <ip> <total time> <idle time> <url>	
void dlr_server( CLIENTBUFFER *client, char *request )
{
	LINKDATA *linkdata;
	char *param;
	time_t now;
	static int init;

	// new, but deprecated until soemeone can tell me what it's good for ...
	if (dlr_sigusr2) {
		time_t thetime;
		// dlr_event("SIGUSR2 - mail check\n");
		time(&thetime);
		if (dial.state == DIALLED) {
			dlr_dorefresh(thetime);
		}
		else {
			// Maybe it's time to do something? Check again!
			dlr_dowakeoff(thetime);
		}
		dlr_sigusr2 = 0;
				
	}
	if (dlr_sigusr1) {
		// dlr_event("SIGUSR1 - Dialling\n");
		dlr_dialnow(NULL,"signalled");
		dlr_sigusr1 = 0;
	}
	if (dlr_sighup) {
		// dlr_event("SIGHUP - reloading config file\n");
		if (dlr_reloadconfig()) {
			dlr_event("Failed: %s\n",strerror(errno));
		}
		else {
			dlr_event("Configuration file reloaded\n");
		}
		pass_reloadfile();
		dlr_sighup = 0;
	}
	if (!init) {
		init = 1;
		atexit(dlr_serverexit);
		dlr_serverstart();
	}

	// We will probably need to know what's going on - so check
	if (dlr_checkstate()) 
		if (!client) return;

	// Split the command and the parameter
	param=strchr(request,' ');
	if (param) 
		*param++ = 0;
	else 
		param="";
	if (strcmp(request,"")==0) {
		// Do nothing -- this is periodically triggered by the server
		// function with a NULL client reference.
	}
	else if (strcasecmp(request,"QUIET")==0) {
		// Switch to quiet mode by closing stdout and stderr
		fclose(stdout);
		fclose(stderr);
		stdout = fopen("/dev/null","w");
		stderr = fdopen(dup2(1,2),"w");
	}
	else if (strcasecmp(request,"HELP")==0) {
		srv_printf(client,
			"LOGIN user pass: authorise my actions\r\n"
			"UID uid key: authorise using key\r\n"
			"START ip: connect this IP on my account (dial if necessary)\r\n"
			"DIAGNOSE: see why\r\n"
			"STOP ip: logout, and hangup if the last\r\n"
			"STATUS ip: show status for IP address\r\n"
			"WHO: which ip's are connected\r\n"
			"HISTORY: graph of history\r\n"
			"LOG: show the last few log lines\r\n"
			// squid: "AUTH url ip/fqdn ident: evaluate URL for link\r\n"
			);
	}
	else if (strcasecmp(request,"STATUS")==0) {
		// STATUS ip
		linkdata = dlr_findlink(param,client,0);
		dlr_showstatus(client,linkdata);
	}
	else if (strcasecmp(request,"LOG")==0) {
		char *p, *eol;
		
		p=dlr_logtail(2000);
		// Strip trailing CRLF so we don't confuse the issue
		eol = p+strlen(p)-1;
		while (eol>p && *(eol-1) <= ' ') {
			eol--;
			*eol=0;
		}
		if (p)
			srv_printf(client,"%s\r\n",p);
		else
			srv_printf(client,"Oops, no log");
		free(p);
	}
	else if (strcasecmp(request,"WHO")==0) {
		linkdata = dlr_findlink(param,client,0);
		dlr_showstatus(client,linkdata);
		dlr_showwho(client);
	}
	else if (strcasecmp(request,"DIAGNOSE")==0) {
		dlr_showdiagnosis(client);
	}
	else if (strcasecmp(request,"HISTORY")==0) {
		dlr_showhistory(client,atoi(param));
	}
	else if (strcasecmp(request,"LOGIN")==0) {
		// LOGIN user pass
		if (dlr_authuser(param,client->userdata,sizeof(client->userdata))) {
			srv_printf(client,"Hello %s from %s\r\n",
				client->userdata, client->peer);
			client -> userflags |= FLAG_AUTH;
		}
		else {
			srv_printf(client,"Error - bad username or password\r\n",param);
			client -> userflags &= ( ! FLAG_AUTH );
		}
	}
	else if (strcasecmp(request,"UID")==0) {
		// UID
		if (dlr_authuid(param,client->userdata,
			sizeof(client->userdata))) {
			// We don't acknowledge this command :)
			//srv_printf(client,"Hello %s from %s\r\n",
			//	client->userdata, client->peer);
			client -> userflags |= FLAG_AUTH;
		}
	}
#if ALLOW_SHUTDOWN
	else if ((client->userflags & FLAG_AUTH) && strcasecmp(request,"SHUTDOWN")==0) {
		dlr_event("Shutdown by %s\n", linkdata->owner);
		exit(1);
	}
#endif
	else if ((client->userflags & FLAG_AUTH) && strcasecmp(request,"LOCK")==0) {
		// LOCK
		dlr_startlink(time(&now),"127.0.0.2",client);
	}
	else if ((client->userflags & FLAG_AUTH) && strcasecmp(request,"UNLOCK")==0) {
		// UNLOCK
		dlr_stoplink(time(&now),"127.0.0.2",client);
	}
	else if ((client->userflags & FLAG_AUTH) && strcasecmp(request,"START")==0) {
		// START ip
		dlr_startlink(time(&now),param,client);
	}
	else if ((client->userflags & FLAG_AUTH) && strcasecmp(request,"STOP")==0) {
		// STOP ip
		dlr_stoplink(time(&now),param,client);
	}
	else {
		srv_printf(client,"Command %s not recognised\r\n",request);
	}
	if (strcmp(request,"")!=0) {
		srv_printf(client,"\r\n");
	}
	return;
}


