// movingavg.c -- Moving average filter `object'
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//
// Well, it's a finite impulse response filter -- but with only one function is
// supported -- a block of 1's.  Who can suggest the equivalent parameters for
// the equivalent butterworth IIR filter?  I don't want to make it too
// computationally intensive though ... :)
//

#include "common.h"
#include "movingavg.h"

// Reset the moving average value
void mva_start(MOVINGAVERAGE *avg, int initialvalue)
{
	memset(avg,0,sizeof(MOVINGAVERAGE));
	mva_addvalue(avg, initialvalue);
}

// How many entries are there in the table
int mva_count(MOVINGAVERAGE *avg)
{
	return avg->count;
}

// How many packets were sent from [t-a] to [t-b] seconds
// a and b range from 1 to count
int mva_getdiff(MOVINGAVERAGE *avg, int a, int b)
{
	if (a>avg->count) a=avg->count;
	if (b>avg->count) b=avg->count;
	a=(avg->pointer-a+MOVINGAVERAGESIZE) % MOVINGAVERAGESIZE;
	b=(avg->pointer-b+MOVINGAVERAGESIZE) % MOVINGAVERAGESIZE;
	return (avg->x[a] - avg->x[b]);
}

// Add a new value to the moving average table
// The value which must be supplied is the running total of the function
// of interest.
void mva_addvalue(MOVINGAVERAGE *avg, int value)
{
	int lastvalue;
	int lastpointer;
	if (avg->count) {
		lastpointer = (avg->pointer - 1 + MOVINGAVERAGESIZE) % MOVINGAVERAGESIZE;
		lastvalue = avg->x[lastpointer];
		if (value+avg->offset < lastvalue) {
			// The value has DECREASED -- so we change the offset 
			// so that x[current(modified)]-x[last] = 0
			avg->offset = lastvalue - value;
			// FIXME -- maybe we should add a relatively large fixed 
			// value, so that we don't hang up after an automatic 
			// redial for a dropped line ... ?  Actually, we should
			// tell someone who cares ... 
		}
		else {
			avg->integral += ( value - lastvalue );
		}
	}
	avg->x[avg->pointer] = value + avg->offset;
	avg->pointer = (avg->pointer+1) % MOVINGAVERAGESIZE;
	if (avg->count < MOVINGAVERAGESIZE)
		avg->count++;
}


// Determine the number of packets over the last n values
// Returns a negative value if:
//    The ticks is more than the data available
// Returns 0 if successful, -1 if failed
// If strict, will return -1, otherwise returns a best-guess
int mva_integral(MOVINGAVERAGE *avg, int ticks, INTEGRAL_TYPE isstrict)
{
	int a,b;
	int total;
	
	// Special case - return the whole bang shoot since the average record
	// was started.  Yep, it wraps for > 2Gb, and yes, it becomes negative
	// Deal with it
	if (isstrict==INTEGRAL_ALL)
		return avg->integral;
	// Too long a period requested?  Well, too bad.  You'll just have
	// to settle for what we've got
	if (ticks > MOVINGAVERAGESIZE ) ticks = MOVINGAVERAGESIZE;
	if (ticks < 2) ticks = 2;
	// Looking for more time than we know about?  Too bad, we can't do
	// this yet -- try again in a few ticks time ...
	if (isstrict==INTEGRAL_STRICT) {
		if (ticks > avg->count) return -1;
	}
	else {
		if (ticks > avg->count) ticks = avg->count;
		if (ticks<2) return 0;
	}
	b = (avg->pointer -   1   + MOVINGAVERAGESIZE) % MOVINGAVERAGESIZE;
	a = (avg->pointer - ticks + MOVINGAVERAGESIZE) % MOVINGAVERAGESIZE;
	total = avg->x[b] - avg->x[a];
	return total;
}

