// shell.c -- command execution and monitoring
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
#include "common.h"

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#include "util.h"
#include "shell.h"

extern char **environ;

// Run a different command
int shell_exec(char *command)
{
	int a;
	char *argv[4];
	argv[0] = "/bin/sh";
	argv[1] = "-c";
	argv[2] = command;
	argv[3] = NULL;
	// FIXME: there a function to do this ...
	for (a = 3; a < 255; a++) // Close like everything except stdin,out
		close(a);
	execve("/bin/sh", argv, environ);
	exit(127); // oops
}

// Do a command in the background
// Returns a non-zero PID on 'success', and 0 on failure
//    prompt: what we display if we fail
//   command: the command to execute in the background
pid_t shell_exec_bg(char *prompt, char *command)
{
	pid_t pid;
	if (!command || !command[0]) return 0; // Nothing to do, no process
	if ((pid=fork())==0) {
		shell_exec(command);
	}
	if (pid<=0) { perror("fork"); 
		pid=0;
	}
	return pid;
}

// Black out dangerous stuff
char *shell_removeescapes(char *param)
{
	char *p;
	p=param;
	while (p && *p) {
		if ( // What we allow
			(*p>='0' && *p<='9') ||
			(*p>='a' && *p<='z') ||
			(*p>='A' && *p<='Z') ||
			(*p=='.') ||
			(*p=='-') ||
			(*p=='_') ||
			(*p==':') ||
			(*p==' ')
			) 
		{
			p++;
		}
		else // How we deny
			*p++ = 'X'; // black it out
	}
	return param;
}

// Read a file into an expanding buffer
char *shell_allocandreadfile(FILE *fd, int buflen, int buflimit)
{
	char *buffer;
	int bytesread,space,pointer;
	
	buffer=xmalloc(buflen);
	bytesread=0;
	pointer=0;
	while (1) {
		space=buflen-pointer-1;
		bytesread = fread(buffer+pointer,1,space,fd);
		if (bytesread>0) {
			pointer+=bytesread;
		}
		buffer[pointer]=0;
		if (ferror(fd)) {
			goto oops;
		}
		if (bytesread==space) {
			// We used it all up
			if (buflen==buflimit) {
				return buffer;  // chop silently - we have hit the roof
			}
			buflen *= 2;
			if (buflen > buflimit) buflen=buflimit;
			buffer=xrealloc(buffer,buflen);
		}
		else {
			break;
		}
	} 
	return buffer;
oops:
	snprintf(buffer,buflen,"%s",strerror(errno));
	return buffer;
}

typedef struct _commandandresult {
	struct _commandandresult *next;
	time_t last;
	char *command;
	char *result;
} COMMANDANDRESULT;

// Execute a command, and return a newly allocated buffer containing the 
// entire output as a NULL terminated string (kinda hoping that the output
// doesn't include meaningful nulls) (the output can also be an error message)
char *shell_inline_alloc(char *command, int *exitstatus)
{
	char *buffer;
	int status;
	FILE *fd;
	if (!command) {
		*exitstatus = 0;
		return NULL;
	}
	fd = popen(command,"r");
	if (!fd) {
		if (exitstatus) *exitstatus=-1;
		return xstrdup(strerror(errno));
	}
	// Read about 1K from the program (but you never know)
	buffer=shell_allocandreadfile(fd, 1024, 65536 /* arbitary limit :) */ );
	fflush(fd);
	status = pclose(fd);	// The exit status of the command
	if (exitstatus) *exitstatus=status;
	return buffer;
}


// Run a command, but not too often.
// Returns a pointer which is valid ONLY UNTIL THE NEXT CALL to this 
// function (don't free it!)
char *shell_inline_buffered(char *command, int timeout)
{
	static COMMANDANDRESULT *commands;
	COMMANDANDRESULT *p, **reference;
	time_t now;
	int exitstatus;

	time(&now);
	p=commands;
	reference=&commands;
	while (p) {
		if (strcmp(command,p->command)==0) {
			if (now - p->last >timeout) {
				break;
			}
			return p->result;  // Yipee! A saving!
		}
		// Possibly remove the command as stale and go on to the next
		if (now - p->last > timeout*10) {
			COMMANDANDRESULT *temp=p->next;
			free(p->command);
			free(p->result);
			free(p);
			*reference = temp;
			p=temp;
		}
		else {
			reference=&p->next;
			p=p->next;
		}
	}
	if (!p) {
		// Make a new entry at the head of the list
		p = xcalloc(1,sizeof(COMMANDANDRESULT));
		p->next = commands;
		p->command = xstrdup(command);
	}
	else {
		// Jump the entry to the front of the queue
		*reference = p->next;
		p->next=commands;
		commands=p;
	}
	// command: contains a copy of the command to be executed
	// result: contains either the last result (defunct) or NULL
	p->last=now;
	free(p->result);
	p->result = shell_inline_alloc(command, &exitstatus);
	return p->result;
}

// Execute a command, and return the results in an existing buffer
int shell_inline(char *command, char *buffer, int *buflen)
{
	FILE *fd = popen(command,"r");
	*buffer=0;
	if (!fd) {
		goto oops;
	}
	*buflen = fread(buffer, 1, *buflen-1, fd);
	if (ferror(fd)) {
		goto oops;
	}
	if (*buflen<0) *buflen=0;
	buffer[*buflen]=0;	// Null terminate, in case
	fflush(fd);
	return pclose(fd);	// Return the exit status of the command

oops:
	*buflen = snprintf(buffer,*buflen,"%s",strerror(errno));
	if (fd) { 
		fflush(fd);
		pclose(fd);	// Return the exit status of the command
	}
	return -1;
}

// Send a signal to the specified shell -- returns 0 if OK
int shell_kill(SHELLINFO *shell, int signal)
{
	return kill(shell->pid,signal);
}

// Die! Die! Die!
void shell_panic(char *message)
{
	perror(message);
	exit(15);
}

enum _FILENO { READ, WRITE};

// Executes a command in the background, and return a structure with
// readable and writable file descriptors for its input and output.
SHELLINFO *shell_create(char *command)
{
	SHELLINFO *shell;
	int mystdin, mystdout, mystderr;
	int pipestdin[2],pipestdout[0];
	pid_t pid;

	// Invent some pipes to write to and read from
	pipe(pipestdin);
	pipe(pipestdout);
	mystdin = dup(0);	if (mystdin==-1) shell_panic("dup(0)");
	mystdout = dup(1);	if (mystdin==-1) shell_panic("dup(1)");
	mystderr = dup(2);	if (mystdin==-1) shell_panic("dup(2)");
	close(0);
	close(1);
	close(2);
	dup(pipestdin[READ]);  // 0
	dup(pipestdout[WRITE]); // 1
	dup(pipestdout[WRITE]); // 2
	pid = fork();
	if (pid==-1) 
		shell_panic("fork");	// This message will go awry, since stdout is loose
	else if (pid) {
		// Restore status quo
		close(0);
		close(1);
		close(2);
		dup(mystdin);  // 0
		dup(mystdout); // 1
		dup(mystderr); // 2
		close(mystdin);
		close(mystdout);
		close(mystderr);

		// We're the parent
		shell=xcalloc(1,sizeof(SHELLINFO));
		if (!shell) 
			shell_panic("alloc");
		shell->pid = pid;
		shell->stdin = pipestdin[WRITE];
		shell->stdout = pipestdout[READ];
		return shell;
	} 
	else {
		// Do the command
		execl("/bin/sh","-c",command,NULL);
		perror("exec");
		exit(15);
	}
	return 0;
}

// Close a connection to a sub-shell
void shell_close(SHELLINFO *shell)
{
	close(shell->stdin);
	close(shell->stdout);
	// And forget about the process :) /?
	// wait for exit?? -- naah, we don't wanna block
}

// Pipe commands
// Open a command pipe, and read its output into a buffer
int pipe_newcommand(char *command, 
		PIPEINFO *pi)
{
	pi->pipe = popen(command,"r");
	if (pi->pipe == NULL) return 1;
	return 0;
}

#if 0
// Read data from a pipe, keeping the most recently read line in the buffer
// A maximum of bufmaximum/2 is garuanteed to be read from the buffer
int pipe_readdata(PIPEINFO *pi)
{
	int margin = pi->bufmaximum/3;
	int space, bytesread;

	// Do we need to rotate the buffer? Yes if there is under a third left
	if (pi->bufmaximum-pi->bufsize < margin) {
		pi->bufsize-=margin;
		memmove(pi->data, pi->data+margin, pi->bufsize);
	}
	// Now read from the process again
	space = pi->bufmaximum-pi->bufsize-1;
	bytesread = fread(pi->data+pi->bufsize, 1, space, pi->pipe);
	if (bytesread<0 || ferror(pi->pipe)) {
	}
	else {

	}
}
#endif



